require "my_util"
local compressor= require "lib/compressor"

local STRING_SIZE=200
local PREFIX="zzzzz"
local DUMMY_NAME=my"dummy-fake-text-data"
local DUMMY_NAME2=my"another-dummy-fake-text-data"

if not (data and data.raw) then 
    return {--I reuse these in scripts/drones.lua
        STRING_SIZE=STRING_SIZE,
        PREFIX=PREFIX,
        DUMMY_NAME=DUMMY_NAME,
        DUMMY_NAME2=DUMMY_NAME2,
    }
end

--these are all different connections between all different types of prototypes
--currently they are made to be one-to-one correspondence
--it would be nice to refactor the code below so that it could handle multiple types of capsules could be used for single robot spawning, and multiple types of robots spawned by a single capsule
local robots,capsules,projectiles,recipes={},{},{},{}
local proj2rob,rob2proj,caps2rob,rob2caps,rob2rec={},{},{},{},{}
local rec2tec,technologies={},{}
local robots_spawned_at_once={}

for n,p in pairs(raw['combat-robot']) do robots[n]=p end

for n,p in pairs(raw.projectile) do
    --print(n)
    local h=has(p,{action={action_delivery={target_effects={type="create-entity"}}}})
    
    local rn=h.entity_name
    --print("creates",rn)
    if robots[rn] then
        --print("corresponds to",robots[rn])
        projectiles[n]=p
        proj2rob[n]=rn
        h.trigger_created_entity=true--for drone life
        
        local offsets=h.offsets
        robots_spawned_at_once[rn]=offsets and #offsets or 1
    end
end

--Debug here
--print("\n-")
--for k in pairs(proj2rob) do print(k) end
--print("\n-")

for n,p in pairs(raw.capsule) do
    local pn= has(p,{capsule_action={attack_parameters={ammo_type={action={action_delivery={type="projectile"}}}}}}).projectile
    --print(serpent.block(p))
    --print('has projectile '..pn)
    if projectiles[pn] then
        capsules[n]=p
        caps2rob[n]=proj2rob[pn]
        --print'found something'
    end
end

--print("Found capsules \n-")
--for k in pairs(capsules) do print(k) end
--print("\n-")

for n,p in pairs(raw.recipe) do 
    if capsules[p.result] then 
        recipes[n]=p
        rob2rec[caps2rob[p.result]]=n
    end 
end
for n,p in pairs(raw.technology) do
    local rn=has(p,{effects={type="unlock-recipe"}}).recipe
    if recipes[rn] then
        rec2tec[rn]=n
        technologies[n]=p
    end
end

local swap=function(a) local b={} for k,v in pairs(a) do b[v]=b[v] or k end return b end
rob2caps=swap(caps2rob)
rob2proj=swap(proj2rob)

local spent=function(text) return "spent#-"..my(text) end
local wrecked=function(text) return "wrecked#-"..my(text) end
local recharge=function(text) return "recharge#-"..my(text)  end
local restore=function(text) return "restore#-"..my(text)  end
local deploying=function(text) return "deploying#-"..my(text) end

local icons=require "data_u/pregenerated_icons"
--local spenticon=function(name) return "__DroneCommand__/graphics/spent_"..name  end
--local wreckedicon=function(name) return "__DroneCommand__/graphics/wrecked_"..name  end

local subgroup1="drone-recharge"
add(derivative(raw["item-subgroup"]["capsule"],{
    name=subgroup1,
    order="cc1",
}))
local subgroup2="drone-restoration"
add(derivative(raw["item-subgroup"]["capsule"],{
    name=subgroup2,
    order="cc2",
}))

for n,robot in pairs(robots) do
    local capsule=capsules[rob2caps[n]]
    local recipe=recipes[rob2rec[n]]
    local projectile=projectiles[rob2proj[n]]
    local technology=technologies[rec2tec[rob2rec[n]]]
    
    if not capsule then error(debug.traceback()) end
    --items
    add(derivative(
        raw.item["iron-plate"],
        {name=spent(n),stack_size=50,icon=icons.spent[n],subgroup=subgroup1}
    ))
    add(derivative(
        raw.item["iron-plate"],
        {name=wrecked(n),stack_size=25,icon=icons.wrecked[n],subgroup=subgroup2}
    ))
    
    
    --recipes
    local count=assert(robots_spawned_at_once[n])
    
    local recipe_recharge=add(derivative(
        derivative(recipe,{ingredients=delet}),
        {
            name=recharge(n),
            ingredients={
                {spent(n),count},
                {"battery",count}
            },
            icon=icons.recharge(n),
            subgroup=subgroup1,
        }
    ))
    
    local drones_per_recipe = (recipe.result_count or 1)*count
    --take 4 reactions worth of wrecked drones and rebuild them for half the price of the same amount of new ones:
    local ingredients={}
    for k,v in pairs(recipe.ingredients) do
        if tonumber(v[2]) or (v[2]==nil) then
            table.insert(ingredients,{v[1],(v[2] or 1) *2})
        else
            log("DroneCommand: Warning: Strange recipe: " .. recipe.name)
        end
    end
    table.insert(ingredients,{wrecked(n),drones_per_recipe*4})
    
    local result_count= drones_per_recipe * 4
    
    local recipe_repair=add(derivative(
        derivative(recipe,{ingredients=delet}),
        {
            name=restore(n),
            ingredients=ingredients,
            result_count=result_count,
            icon=icons.restore(n),
            subgroup=subgroup2,
        }
    ))
    
    --projectiles
    add(derivative(projectile,{name=deploying(n),action=delet}))
    
    --update technology
    table.insert(technology.effects,{type="unlock-recipe",recipe=recipe_recharge.name})
    table.insert(technology.effects,{type="unlock-recipe",recipe=recipe_repair.name})
end

--passing data to control.lua
local compressed_data=compressor.compress(rob2caps)
local compressed_data2=compressor.compress(robots_spawned_at_once)--this all could use a refactoring, for now it sticks to the code that is working

local generate_dummy_entity=function(ent_name,text)
    local cp,ind=1,1
    while #text>0 do
        local part=text:sub(cp,cp+STRING_SIZE-1-#PREFIX)
        --print("BEFORE C")
        --print(part)
        add(derivative(raw["flying-text"]["flying-text"],{name=ent_name..ind,
            order=PREFIX..part
        }))
        cp=cp+STRING_SIZE-#PREFIX
        text=text:sub(cp)
        ind=ind+1
    end
end
generate_dummy_entity(DUMMY_NAME,compressed_data)
generate_dummy_entity(DUMMY_NAME2,compressed_data2)
