--[[Handles the management of gui objects
Most notably, handles the events.
]]
local gui=require "lib/gui"
local roots={top=true,left=true,center=true}
local gui_dispatch
gui_dispatch={
	create=function()--create dispatcher
		local d={}
		for k,v in pairs(gui_dispatch.__index) do
			d[k]=v
		end
		return d
	end,
	------
	__index={
		attach=function(me,handle)--attach global storage for the dispatcher and restore the metatables of guis
			me.data=handle
			for _,player in pairs(handle) do 
				for _,g in pairs(player) do 
					gui.restore(g) 
				end
			end
		end,
		get_events=function() return {--shortcut to event codes, that are of interest for the dispatcher
			defines.events.on_gui_checked_state_changed,
			defines.events.on_gui_click,
			defines.events.on_gui_text_changed,
			}
		end,
		get_handler=function(me) return function(event) me:on_gui_event(event)  end  end,
		on_gui_event=function(me,event)--should be fed  to script on_event
			local pind=event.player_index
			if me.data[pind] then
				for k,g in pairs(me.data[pind]) do 
					if g.anchor.valid then 
						g:check(event.element) 
					else
						me.data[pind][k]=nil
					end
				end
			end
		end,
		open=function(me,parameters)--creates the root element of a gui and starts gui growth
			--[[ 
			parameters={
				player_index,
				name,--!mandatory
				root,--!mandatory
				type,--!mandatory
				style,
				data,
				on_close,
				on_duplicate,
				elements,
				children,
				}
				this table is rethrown downstream to gui.new() function
			]]
			--local me=gui_dispatch
			local pind=parameters.player_index
			local name=parameters.name
			me.data[pind]=me.data[pind] or {}
			
			::retry::--this is label for goto
			if me.data[pind][name] and me.data[pind][name].anchor.valid then
				if me.data[pind][name].on_duplicate then 
					me.data[pind][name]:on_duplicate()--guis can handle teir duplicates gracefully
					goto retry--this is the goto, no other way came to mind with this 
				else
					error  ('gui element already present'..name)
				end
				return 
			else--if no duplicates
				local p=parameters
				if not roots[p.root] then error ('incorrect gui location specified: '..serpent.block(p.root) .. debug.traceback()) end
				--p.anchor=game.players[pind].gui[p.root].add{name=p.name,type=p.type,style=p.style}
				p.anchor=game.players[pind].gui[p.root].add{name=p.name,type=p.type,style=p.style}
				me.data[pind][name]=gui.new(p)
				return me.data[pind][name]
			end
		end,
	}
}
return gui_dispatch
