--[[specification of the class to facilitate fast lookup and arbitrary traversing on discontinuous keys
usage:
    list_type_generator = require("hashed_list")
    list_type=list_type_generator(function(a,b) return a<b end )
    list=list_type.new(elements)
    
    list_type.restore(list)
]]

local link=function(key,value)
    return {
        left=link,
        right=link,
        key=key,
        value=value,
    }
end

local bubble_down
bubble_down=function(order_fun,cur,new)
    --order_fun --return true if cur < new
    if order_fun(cur.value,new.value) then
        local l,r=cur,cur.right
        if l then l.right=new end
        if r then r.left=new end
        new.right=r
        new.left=l
    else
        if cur.left then
            bubble_down(order_fun,cur.left,new)
        else
            cur.left=new
            new.right=cur
        end
    end
end

local function list_type_generator(order_fun)
    local list
    list={
        new=function(elements)
            local l = setmetatable({hash={},n=0,left=nil,right=nil},list)
            for k,v in pairs(elements or {}) do
                l:add(k,v)
            end
            return l
        end,
        restore=function(t)
            return setmetatable(t,list)
        end,
        __index={        
            add=function(me,k,v)
                local l = link(k,v)
                if me.hash[k] then me:remove(k) end
                me.hash[k]= l
                me.n=me.n+1
                if me.n==1 then 
                    me.left=l me.right=l 
                else
                    bubble_down(order_fun,me.right,l)
                    if me.right==l.left then me.right=l end
                end
            end,
            remove=function(me,k)
                if not me.hash[k] then return  end
                me.n=me.n-1
                local v=me.hash[k]
                me.hash[k]=nil
                local l,r=v.left,v.right
                if l then 
                    l.right=r 
                else
                    me.left=r
                end
                if r then 
                    r.left=l 
                else
                    me.right=l
                end
            end,
            get=function(me,k)
                local c=me.hash[k]
                return c and c.value
            end,
            previous=function(me,k)
                local c = me.hash[k]
                local t = c and c.left or me.right
                if t then return t.key, t.value end
            end,
            next=function(me,k)
                local c = me.hash[k]
                local t = c and c.right or me.left
                if t then return t.key, t.value end
            end,
        }
    }
    return list
end

return list_type_generator
