local dispatch_creator=require "lib/gui_dispatch"
local list_type_generator=require "lib/hashed_list"

local gd=dispatch_creator.create()
local data

local list=list_type_generator(function(a,b) return a.signal < b.signal end)
local newlist=function() return list.new() end
local restorelist=function(_,_,t) list.restore(t)  end

on_init(function()
    gd:attach(new_storage("label-gui"))
    data=ad_hoc_table(new_storage("label"),newlist,restorelist)
end)
on_load(function()
    gd:attach(get_storage("label-gui"))
    data=ad_hoc_table(get_storage("label"),newlist,restorelist)
end)

local isnum=function(variable) return type(variable)=="number"  end
local isstring=function(variable) return type(variable)=="string"  end
local set_label=function(position,signal,forceid,playerid)
    assert(isnum(signal))
    local storage=data[forceid]
    local prev=table.deepcopy(storage:get(signal))
    storage:add(signal,{
        signal=signal,
        position=position,
        set_by=playerid,
        previous=prev and (not (prev.set_by==playerid) and prev.position or prev.previous)
        })
end

local get_label=function(forceid,signal)
    assert(isstring(forceid))
    assert(isnum(signal))
    return data[forceid]:get(signal)
end

local del_label=function(name,forceid,playerid)
    local storage = data[forceid]
    local l = storage:get(name)
    if l then
        --if not (l.set_by==playerid) then
            --l.previous=l.position
            --l.set_by=playerid
        --end
        --l.deleted=true
        storage:remove(name)
    end
end
    
local next_label=function(forceid,current) return data[forceid]:next(current ) end
local prev_label=function(forceid,current) return data[forceid]:previous(current ) end

--a set of global functions that can be used by other part of the mod
define{label={set=set_label, get=get_label, del=del_label, next=next_label, prev=prev_label}}

on_gui(gd:get_handler())

on_label_key(function(e)
    --print(game.players[e.player_index].gui.center.name)
    gd:open{
        player_index=e.player_index,
        name="label-gui",
        root="center",
        type="frame",
        orientation="vertical",
        caption="label-gui",
        on_duplicate=function(gui) gui:destroy() end,
        children={
            {type="flow",
            name="label-display",
            orientation="horizontal",
            elements={
                {type='label',caption="x: ",},
                {type="textfield",name='x',},
                {type='label',caption="y: ",},
                {type="textfield",name="y",},
                {type='label',name="namelabel",caption="name:",},
                {type="textfield",name="nametext",},
            }},
            {type="flow",
            name="control",
            orientation="horizontal",
            data={
                init=function(g)  
                    if g.data.initted then return end
                    g.data.initted=true
                    local root=g:get_top()
                    g.data.root=root
                    local fields=root:find_down{type="textfield"}
                    --print(serpent.block(fields))
                    for k,v in pairs(fields) do
                        g.data[v.name]=v
                    end
                    g.data.output=root.elements.output
                end,
                initted=false,
                get_input=function(me)
                    --print(serpent.block(me))
                    local x=tonumber(me.x.text) 
                    local y=tonumber(me.y.text) 
                    local name=tonumber(me.nametext.text) 
                    
                    if not (x and y and name) then
                        me:set_input(x or 0, y or 0, name or 0)
                        me:set_output("gui-incorrect-input")
                        return 
                    end
                    return x,y,name
                end,
                set_input=function(me,x,y,name)
                    me.x.text=x 
                    me.y.text=y 
                    me.nametext.text = name                         
                end,
                set_output=function(me,output)
                    me.output.caption=output
                end,
            },
            elements={
                {type="button",name="set",caption="set",
                    handler=function(g,b)
                        local data=g.data
                        data.init(g)
                        local x,y,name=data:get_input()
                        if (not (x and y and name)) or name==0 then 
                            data:set_output"gui-label-incorrect-inputs" 
                            return 
                        end
                        local pind=g.anchor.player_index
                        local p=game.players[pind]
                        local prev=label.get(p.force.name,name)
                        if prev then
                            data:set_output"gui-label-changed"
                        else
                            data:set_output"gui-label-created"
                        end
                        label.set({x=x,y=y},name,p.force.name,pind)
                    end
                },
                {type="button",name="remove",caption="remove",
                    handler=function(g,b)
                        local data=g.data
                        data.init(g)
                        local x,y,name=data:get_input()
                        if not x then return  end
                        local pind=g.anchor.player_index
                        local p=game.players[pind]
                        
                        local prev=label.get(p.force.name,name)
                        if prev and not (prev.position.x==x and prev.position.y==y) then
                            data:set_input(prev.position.x,prev.position.y,name)
                            data:set_output"gui-label-repeat"
                            return
                        else
                            label.del(name,p.force.name,pind)
                            data:set_input("","","")
                            data:set_output"gui-label-deleted"
                        end
                            
                    end
                },
                {type="button",name="next",caption="next",
                    handler=function(g,b)
                        local data=g.data
                        data.init(g)
                        local _,_,name=data:get_input()
                        local fid=game.players[g.anchor.player_index].force.name
                        local n=label.next(fid,name)
                        local l=label.get(fid,n)
                        data:set_input(l.position.x,l.position.y,n)
                        data:set_output("")
                    end
                },
                {type="button",name="previous",caption="previous",
                    handler=function(g,b)
                        local data=g.data
                        data.init(g)
                        local _,_,name=data:get_input()
                        local fid=game.players[g.anchor.player_index].force.name
                        local n=label.prev(fid,name)
                        local l=label.get(fid,n)
                        data:set_input(l.position.x,l.position.y,n)
                        data:set_output("")
                    end
                },
            }},
        },
        elements={
            {type="label",name="output",caption=""},
            {type="button",name="close",caption="close",handler=function(me) me:destroy() end},
        }
    }
end)
