require "my_util"
require "scripts/drones"

local DRONES

local DUMMY_ITEM=my"command-transmitter-item"

local LANDING_RANGE=30
local LANDING_RANGE2=LANDING_RANGE^2
local LANDING_PAD_NAMES={
    [my"basic-landing-pad"]={
        spent={lightly_damaged=0.5,heavily_damaged=0.5},
        killed={lightly_damaged=0.3,heavily_damaged=0.7},
    },
    [my"powered-landing-pad-inventory"]={
        powered=1e4,--J
        spent={lightly_damaged=0.9,heavily_damaged=0.1},
        killed={lightly_damaged=0.6,heavily_damaged=0.4},
    },
    [my"powered-landing-pad-power"]={
        powered=true,
        inventory=my"powered-landing-pad-inventory",
    }
}
    local rand=math.random
local is_light_damage=function(pad,is_killed)
    local t=LANDING_PAD_NAMES[pad.name]
    local c= is_killed and t.killed or t.spent
    return (rand()<c.lightly_damaged)
end
local is_powered=function(name)
    return LANDING_PAD_NAMES[name].powered
end

local stack
stack={
    --container with self-caching search
    new=function()
        local o={l=1,r=0}
        for n,f in pairs(stack) do
            o[n]=f
        end
        return o
    end,
    add=function(me,el)
        me.r=me.r+1
        me[me.r]=el
    end,
    search=function(me,func)
        local l,r=me.l,me.r
        if l>r then return end
        if func(me[l]) then return me[l] end
        for i=l+1,r do
            local element=me[i]
            if func(element) then
                me.l=l-1
                me[me.l]=element
                me[i]=me[r]
                me[r]=nil
                me.r=r-1
                return element
            end
        end
    end,
    remove=function(me,func)
        local l,r=me.l,me.r
        for i=r,l,-1 do
            if func(me[i]) then
                me[i]=me[r]
                me[r]=nil
                me.r=r-1
            end
        end
    end,
    search_and_remove=function(me,func)
        local l,r=me.l,me.r
        if l>r then return  end
        local crit=func(me[l])
        if crit==1 then return me[l] end
        local i=l+1
        while i<=r do
            local element=me[i]
            local crit=func(element)
            if crit==1 then
                me.l=l-1
                me[me.l]=element
                me[i]=me[r]
                me[r]=nil
                me.r=r-1
                return element
            elseif crit==-1 then
                me[i]=me[r]
                me[r]=nil
                r=r-1
                me.r=r
            else
               i=i+1 
            end
        end
    end
}

local double_dynamic=function(t)
    return setmetatable(t,{__index=function(t,k)
        t[k]=setmetatable({},{__index=function(tt,kk)
            --print('asking ',k,kk)
            tt[kk]=stack.new()
            return tt[kk]
        end})
        return t[k]
    end})
end

local DATA,IGNORED
on_init(function()
    DATA=double_dynamic(new_storage("landing-pads"))
    DRONES=get_storage("drone-types")
    IGNORED=new_storage('ignored-landings')
end)
on_load(function()
    DATA=double_dynamic(get_storage("landing-pads"))
    DRONES=get_storage("drone-types")
    IGNORED=get_storage('ignored-landings') 
end)

    
on_build(function(event)
    local e=event.created_entity
    if not LANDING_PAD_NAMES[e.name] then return end
    local f=e.force.name
    local s=e.surface.index
    local pads=DATA[s][f]
    if is_powered(e.name) then
        local pps=LANDING_PAD_NAMES
        local name=pps[e.name].inventory
        local inv=game.surfaces[s].create_entity{name=name,position=e.position,force=f}
        pads:add{entity=inv,power_entity=e,powered=pps[name].powered,name=inv.name,last_landed=DUMMY_ITEM}
    else
        pads:add{entity=e,powered=false,name=e.name,last_landed=DUMMY_ITEM}
    end
end)

local is_near, set_near
do
    local pos1
    set_near=function(pos) pos1={x= pos.x or pos[1],y=pos.y or pos[2]}  end
    is_near=function(pos)
        local x = pos.x or pos[1]
        local y = pos.y or pos[2]
        return ((pos1.x-x)^2+(pos1.y-y)^2)<LANDING_RANGE2
    end
end
local can_accept=function(pad)
    local e=pad.entity
    local ppe=pad.power_entity
    if (not e.valid) then 
        if ppe and ppe.valid then ppe.destroy() end
        return -1
    end
    if (ppe and ppe.energy<pad.powered) then return end
    return is_near(e.position) and e.can_insert(pad.last_landed) and 1
    --there will be some mishaps with almost-full landings but whatever
end
local find_pad=function(pos,surf,force)  
    local st=DATA[surf.index][force.name]
    set_near(pos)
    return st:search_and_remove(can_accept)
end

local this_ent=function(entity)
    return 
    function(pad)  
        local ent,pow=pad.entity,pad.power_entity
        if not ent.valid or (ent==entity) then return true end 
        if pow and (not pow.valid or pow==entity) then ent.destroy() return true end
    end
end
local handle_removal=function(ent)
    if not LANDING_PAD_NAMES[ent.name] then return end
    local surf,force=ent.surface,ent.force
    local st=DATA[surf.index][force.name]
    st:remove(this_ent(ent))
end
local ignore_death=function(entity)
    if entity.valid then IGNORED[entity.unit_number]=true end
end
local should_ignore=function(entity)
    if entity.valid then 
        local i=entity.unit_number
        if IGNORED[i] then
            IGNORED[i]=nil
            return true
        end
    end
end
on_die(function(event)
    local e=event.entity
    handle_removal(e)
    if not DRONES[e.name] then return end
    if should_ignore(e) then return end
    --print(e.name,DRONES[e.name])
    local pos,surf,force=e.position,e.surface,e.force
    local pad=find_pad(pos,surf,force)
    if not pad then return end
    assert(pad.entity.valid)
    --print(serpent.block(pad))
    local D=DRONES[e.name]
    local killed = not (event.force == game.forces.neutral )
    local item= (is_light_damage(pad,killed) and D.lightly_damaged) or D.heavily_damaged
    --print(item)
    local projectile= D.projectile
    local pad_ent=pad.entity
    pad_ent.insert({name=item,count=1})
    pad.last_landed=item
    surf.create_entity{name=projectile,position=pos,target=pad_ent,speed=0.4}
    if pad.powered then
        pad.power_entity.energy=pad.power_entity.energy-pad.powered
    end
end)

on_mined(function(event)
    handle_removal(event.entity)
end)

define{LANDING_PADS={find=find_pad,ignore_death=ignore_death}}
return LANDING_PADS
