require "my_util"
require "scripts/label"

local NEUTRAL_FORCE_NAME="neutral"--should be removed when get_cease_fire is fixed

local STEP=60 -- ticks between updates of a given entity
local SCAN_DURATION=15--number of steps to finish the scan
local ENERGY_PER_SCAN=9e5--energy drained per scan
local MAX_RANGE=20*32--maximum scanning range in tiles (chunks*chunk_size)
local MAX_RANGE2=(MAX_RANGE/32)^2

--index of few special output signal
local MAX_RANGE_IDX=1
local OUT_OF_RANGE_IDX=2
local TIME_REMAINING_IDX=3
local X_IDX=4
local Y_IDX=5
local X_OWN=6
local Y_OWN=7
local FREE_IDX=8
local LAST_IDX=18

local input_ent=my"smart-radar-input"
local output_ent=my"smart-radar-output"


local signals={map_label=my"map-label"}
for n,t in pairs(require"data_u/signals") do
    if n:find("radar") then
        local pat=my"radar%-(.*)%"
        for _,s in pairs(t) do
            signals[s:match(pat):gsub("%-","_")]=s
        end
        break
    end
end

local MAX_RANGE_SIGNAL={count=MAX_RANGE,index=MAX_RANGE_IDX,signal={type="virtual",name=signals.max_radius}}
local OUT_OF_RANGE_SIGNAL={count=1,signal={type="virtual",name=signals.out_of_range},index=OUT_OF_RANGE_IDX}

local TIME_REMAINING_ID={type="virtual",name=signals.time_remaining}
local X_SIGNAL_ID={type="virtual",name=signals.posX}
local Y_SIGNAL_ID={type="virtual",name=signals.posY}
local FORCE_OWN_ID={type="virtual",name=signals.force_own}
local FORCE_FRIEND_ID={type="virtual",name=signals.force_friend}
local FORCE_NEUTRAL_ID={type="virtual",name=signals.force_neutral}
local FORCE_ENEMY_ID={type="virtual",name=signals.force_enemy}
local X_OWN_ID={type="virtual",name=signals.ownX}
local Y_OWN_ID={type="virtual",name=signals.ownY}



local get_entities--returns entities on the surface in the area satisfying the diplomacy filters
do
    local make_force_getter=function(method,criterion)
        criterion=(criterion==false) and false or true
        return function(force)
            local r={}
            local n=1
            for _,f in pairs(game.forces) do
                if force[method](f)==criterion then
                    r[n]=f
                    n=n+1
                end
            end
            return r
        end
    end

    local filters={
        force_own=function(f) return f end,
        force_friend=make_force_getter("get_friend"),
        force_neutral=make_force_getter("get_cease_fire"),
        force_enemy=make_force_getter("get_cease_fire",false),
    }
    local append=function(t1,t2) 
        local n=#t1
        for i=1,#t2 do
            t1[n+i]=t2[i]
        end
    end
    get_entities=function(force,surface,area,filter)
        local r={}
        if filter then
            for _,f in pairs(filters[filter](force)) do
                append(r,surface.find_entities_filtered{area=area,force=f})
            end
        else
            r=surface.find_entities(area)
        end
        return r
    end
end

local cache
local entities_to_signals--convert entities to the signals of their mining results
do
    entities_to_signals=function(force,entities)
        local ans,force_count={},{}
        local ent_count,ent_signals={},{}
        local prototypes=game.entity_prototypes
        local forces={[force.name]=FORCE_OWN_ID}
        for i=1,#entities do
            local e=entities[i]
            local n=e.name
            --creation of siganls form entities
            local seen,converted=cache.seen[n],cache.signal[n]
            if not seen then
                cache.seen[n]=true
                local pr=prototypes[n]
                local m=pr.mineable_properties
                if m.minable and m.products and m.products[1] then
                    local p=m.products[1]
                    converted={type=p.type,name=p.name}
                elseif pr.items_to_place_this then
                    local name=next(pr.items_to_place_this) 
                    if name then converted={type ="item", name= name} end
                end
                cache.signal[n]=converted
            end
            if converted then
                --creation of signals of total counts of entities per force
                --only entities with signals count
                local f=e.force
                local stance=forces[f.name]
                if not stance then
                    --print(force.name,f.name)
                    --print(force.get_friend(f))
                    --print(force.get_cease_fire(f))
                    if force.get_friend(f) then forces[f]=FORCE_FRIEND_ID
                    elseif force.get_cease_fire(f) or f.name==NEUTRAL_FORCE_NAME then forces[f]=FORCE_NEUTRAL_ID
                    else forces[f]=FORCE_ENEMY_ID
                    end
                    stance=forces[f]
                end
                ent_count[converted.name]=(ent_count[converted.name] or 0) + 1
                ent_signals[converted.name]=converted
                force_count[stance]=(force_count[stance] or 0) +1
            end
        end
        local s=FREE_IDX
        for id,count in pairs(force_count) do
            ans[#ans+1]={count=count, signal=id, index=s}
            s=s+1
            if s==LAST_IDX then break end
        end
        for name,count in pairs(ent_count) do
            ans[#ans+1]={count=count, signal=ent_signals[name], index=s}
            s=s+1
            if s==LAST_IDX then break end
        end
        return ans
    end
end


local data
on_init(function()
    data=new_storage("smart-radar")
    cache=ad_hoc_table(new_storage("radar-conversion-cache"))
end)
on_load(function()
    data=get_storage("smart-radar")
    cache=ad_hoc_table(get_storage("radar-conversion-cache"))
end)

local GREEN,RED=defines.wire_type.green,defines.wire_type.red
local invalid={valid=false}
local connectable={
    ["arithmetic-combinator"]=defines.circuit_connector_id.combinator_input,
    ["decider-combinator"]=defines.circuit_connector_id.combinator_input,
    ["constant-combinator"]=defines.circuit_connector_id.constant_combinator,
}
local insert=table.insert

local new_radar=function(e)
    local s=e.surface
    local p=e.position
    local f=e.force
    local out = s.create_entity{name=output_ent,position={p.x+1,p.y+1},force=f}
    local output=out.get_or_create_control_behavior()
    --output.set_signal(MAX_RANGE_IDX,MAX_RANGE_SIGNAL)
    output.parameters={parameters={
            MAX_RANGE_SIGNAL,
            OUT_OF_RANGE_SIGNAL,
            {count=p.x,index=X_OWN,signal=X_OWN_ID},
            {count=p.y,index=Y_OWN,signal=Y_OWN_ID},
        }
    }
    local o={
        input_ent=e,
        output_ent=out,
        in_green=invalid,
        in_red=invalid,
        out=output,
        cx=(p.x-p.x%32)/32,
        cy=(p.y-p.y%32)/32,
        surface=e.surface,
        task={},
        }
    insert(data,o)
    --hack to improve blueprint usability
    local area={{p.x,p.y+1.5},{p.x+2,p.y+2.5}}
    for _,c in pairs(s.find_entities_filtered{area=area,force=f}) do
        local id=connectable[c.name]
        if id then
            out.connect_neighbour{
                target_entity=c,
                wire=RED,
                target_circuit_id=id
            }
        end
    end    
end

local old_radar=function(e)
    local p,s,f=e.position,e.surface,e.force
    local area={{p.x-1,p.y-1.5},{p.x+1,p.y-.5}}
    
    for _,c in pairs(s.find_entities_filtered{area=area,force=f}) do
        if c.name==output_ent then
            c.connect_neighbour{
                target_entity=e,
                wire=RED,
                target_circuit_id=connectable[e.name],
            }
        end
    end
end

on_build(function(event)
    local e=event.created_entity
    if (e.name==input_ent) then 
        new_radar(e)
    elseif connectable[e.name] then
        old_radar(e)
    end
end)

local get_signals=function(r)
    --update network connections
    local in_ent,out_ent=r.input_ent,r.output_ent
    if not r.in_green.valid then r.in_green= in_ent.get_circuit_network(GREEN) or invalid end
    if not r.in_red.valid then r.in_red= in_ent.get_circuit_network(RED) or invalid end
    if not r.out.valid then r.out= out_ent.get_or_create_control_behavior() or invalid end
    
    --update input signals
    local in_signals={}
    if r.in_green.valid then
        local green=r.in_green.signals
        if green then
            for i=1,#green do
                local s=green[i]
            --for ,s in pairs(green) do
                in_signals[s.signal.name]=s.count 
            end
        end
    end
    if r.in_red.valid then
        local red=r.in_red.signals
        if red then
            for i=1,#red do
                local s=red[i]
            --for k,v in pairs(r.in_red.signals) do
                local name,count=s.signal.name,s.count
                in_signals[name]=(in_signals[name] or 0)+count
            end
        end
    end
    return in_signals
end

local get_x_y=function(in_signals)
    local inx,iny
    local l=in_signals[signals.map_label]
    if l then
        local pos= label.get(in_ent.force.name,l)
        if pos then
            local p=pos.position
            inx=p.x
            iny=p.y
        end
    else
        inx,iny= in_signals[signals.posX], in_signals[signals.posY]
    end
    return inx,iny
end

local same_order=function(r)
    local task=r.task
    local in_ent=r.input_ent
    if (in_ent.energy> ENERGY_PER_SCAN) then 
        task.time_remaining=task.time_remaining-1
    end
    if task.time_remaining>0 then
        local par=r.last_scan or {}
        insert(par,{count=task.time_remaining,signal=TIME_REMAINING_ID,index=TIME_REMAINING_IDX})
        return par
    else--if tasked scan is done
        in_ent.energy=in_ent.energy-ENERGY_PER_SCAN
        
        local tx,ty=task.x,task.y
        local force=in_ent.force
        local surf=in_ent.surface
        local area={{tx*32,ty*32},{(tx+1)*32,(ty+1)*32}}
        force.chart(surf,{{32*tx+16,32*ty+16},{32*tx+16,32*ty+16}})
    
        local entities=get_entities(force,surf,area,filter)
        local more_parameters=entities_to_signals(force,entities)
        
        r.last_scan=more_parameters
        --prepare new loop
        task.x=nil
        task.y=nil
        return more_parameters
    end
end

local new_order=function(r,x,y)
    local task=r.task
    if (x-r.cx)^2+(y-r.cy)^2<MAX_RANGE2 then
        --if new task
        task.x=x
        task.y=y
        task.time_remaining=SCAN_DURATION
        return r.last_scan or {}
    else
        --maybe should move it so that wrong (x,y) are returned but timer not decreased
        --if out of range
        task.x=nil
        task.y=nil
        task.time_remaining=SCAN_DURATION
        return {OUT_OF_RANGE_SIGNAL,}
    end
end

local update=function(r)
        --check the task
        local in_signals=get_signals(r)
        local inx,iny=get_x_y(in_signals)
        if not (inx and iny) then return end

        --transfer to chunk coordinates
        local x=(inx-(inx%32))/32
        local y=(iny-(iny%32))/32
        --if task is given
        local task=r.task
        local tx,ty=task.x,task.y
        
        local more_parameters
        if (tx==x)and(ty==y) then
            more_parameters=same_order(r)
        else
            more_parameters=new_order(r,x,y)
        end
        local parameters={
            MAX_RANGE_SIGNAL,
            {count=x*32,index=X_IDX,signal=X_SIGNAL_ID},
            {count=y*32,index=Y_IDX,signal=Y_SIGNAL_ID},
            {count=r.cx*32,index=X_OWN,signal=X_OWN_ID},
            {count=r.cy*32,index=Y_OWN,signal=Y_OWN_ID},
        }
        local j=#parameters
        for i=1,#more_parameters do
            parameters[i+j]=more_parameters[i]
        end
        --set the parameters to the output combinators
        r.out.parameters={parameters=parameters}
end

local is_valid=function(r)
    local in_ent, out_ent = r.input_ent, r.output_ent
    if not (in_ent.valid and out_ent.valid ) then
        if in_ent.valid then in_ent.destroy() end
        if out_ent.valid then out_ent.destroy() end
        return false
    end
    return true
end
on_tick(function(event)
    --on tick handler is written in ugliest way for the sake of speed gain however small
    --the function calls do make difference, but noticeable only for tiny primitive functions
    local t=event.tick
    for i=#data-t%STEP,1,-STEP do--update each 60th radar each tick
        local r=data[i]
        if is_valid(r) then
            update(r)
        else
            table.remove(data,i) 
        end
    end
end)
