--generates table of functions that operate on squads to perform special actions
--indexed by the transformed signal names

local CARGO_DRONE_LOAD_RANGE=10--tiles
local CARGO_DRONE_CAPACITY=100--items
local CARGO_DROP_CHEST="logistic-chest-passive-provider"

local CONSTRUCTION_DROP_RANGE=10--tiles
local CONSTRUCTION_DROP_RESULT="roboport"
local CONSTRUCTION_PACKED_ROBOTS1={name="construction-robot",count=10}
local CONSTRUCTION_PACKED_ROBOTS2={name="logistic-robot",count=10}
local CONSTRUCTION_PACKED_PACKS={name="repair-pack",count=20}

local DUMMY_ITEM= my"unexistant-item"

local LANDING_PADS=require "scripts/landing"

local insert=table.insert
local remove=table.remove

local check_valid=function(t)
    for i=#t,1,-1 do if not t[i].entity.valid then
        remove(t,i)
    end end
    return t
end
local get_inventories=function(chests)
    local dic=defines.inventory.chest
    local inventories={}    
    for i=1,#chests do inventories[i]=chests[i].get_inventory(dic) end
    return inventories
end
local squad_info=function(squad)
    local g=squad.guide
    local pos,surf,force=g.position,g.surface,g.force
    local order=squad.order
    return pos,surf,force,order
end
local gen_special_drone_iterator=function(array,move_on)
    --loop for conditional moving on
    --can lead to infinite loops
    local i,N=1,#array
    return function()
        --print('iterator',i,'move',move_on(array[i]))
        while move_on(array[i]) do 
            i=i+1 
            if i>N then return end
        end
        return array[i]
    end
end
local cargo_is_empty=function(drone) return drone.total_count==0 end
local cargo_is_full=function(drone) return drone.total_count==CARGO_DRONE_CAPACITY  end

local inventory_is_empty=function(inventory) return inventory.get_item_count()==0  end    
local inventory_is_full=function(inventory) return not inventory.can_insert(DUMMY_ITEM) end

local valid_construction=function(entities)
    for i=#entities,1,-1 do
        if not entities[i].valid then remove(entities,i) end
    end
    return entities
end
local insert_items=function(port)
    local robots=port.get_inventory(defines.inventory.roboport_robot)
    local wrenches=port.get_inventory(defines.inventory.roboport_material)
    robots.insert(CONSTRUCTION_PACKED_ROBOTS1)
    robots.insert(CONSTRUCTION_PACKED_ROBOTS2)
    wrenches.insert(CONSTRUCTION_PACKED_PACKS)
end
local spawn_roboport=function(drone,surface,position,force)  
    local params={name=CONSTRUCTION_DROP_RESULT,position=position,force=force}
    if not surface.can_place_entity(params) then return false end
    local port=surface.create_entity(params)
    insert_items(port)
    script.raise_event(defines.events.on_robot_built_entity,{
        robot=drone,
        created_entity=port,
        name=defines.events.on_robot_built_entity,
        tick=game.tick,
    })
    LANDING_PADS.ignore_death(drone)
    drone.die()
    return true
end

local SPECIAL_HANDLERS= {
    land=function(squad)
        local pos,surf,force=squad_info(squad)
        if LANDING_PADS.find(pos,surf,force) then
            squad:kill()
            return true
        end
    end,
    load_cargo=function(squad)
        local p,s,f,o=squad_info(squad)
        local pos=o.position
        if not pos then return false end
        if (pos.x-p.x)^2+(pos.y-p.y)^2>CARGO_DRONE_LOAD_RANGE^2 then return false end
        --if ((pos.x-p.x)>CARGO_DRONE_LOAD_RANGE) or ((pos.y-p.y)>CARGO_DRONE_LOAD_RANGE) then return false end
        local chests=s.find_entities_filtered{
            area={{pos.x-1,pos.y-1,},{pos.x+1,pos.y+1}},
            type = "logistic-container",
            force = f
        }
        
        if #chests==0 then return false end
        
        local inventories=get_inventories(chests)
        local drones=check_valid(squad.cargo)
        if #drones==0 then return false end
        
        local next_drone=gen_special_drone_iterator(drones,cargo_is_full)
        local next_inv=gen_special_drone_iterator(inventories,inventory_is_empty)
        
        local dr,inv=next_drone(),next_inv()
        while dr and inv do
            for name,count in pairs(inv.get_contents()) do
                local n=CARGO_DRONE_CAPACITY- dr.total_count
                local taken=(n>0) and inv.remove{name=name,count=n} or 0
                
                dr.total_count=dr.total_count+taken
                dr.items[name]=(dr.items[name] or 0) +taken
            end
            dr,inv=next_drone(),next_inv()
        end
        --print(serpent.block(drones))
        
        return true
    end,
    drop_cargo=function(squad)
        local p,s,f,o=squad_info(squad)
        local pos=o.position
        
        if (not pos) or ((pos.x-p.x)^2+(pos.y-p.y)^2>CARGO_DRONE_LOAD_RANGE^2) then return false end
        --if ((pos.x-p.x)>CARGO_DRONE_LOAD_RANGE) or ((pos.y-p.y)>CARGO_DRONE_LOAD_RANGE) then return false end
        
        local drones=check_valid(squad.cargo)
        local drone=drones[1]
        if not drone then return false end
        
        local params={name=CARGO_DROP_CHEST,position=pos,force=f}
        if not s.can_place_entity(params) then return false end
        local chest=s.create_entity(params)
        local inv=chest.get_inventory(defines.inventory.chest)
        
        for name,count in pairs(drone.items) do if count and count>0 then
            inv.insert{name=name,count=count}
        end end
        
        remove(squad.cargo,1)
        script.raise_event(defines.events.on_robot_built_entity,{
            robot=drone.entity,
            created_entity=chest,
            name=defines.events.on_robot_built_entity,
            tick=game.tick,
        })
        LANDING_PADS.ignore_death(drone.entity)
        drone.entity.die()
        return true
    end,
    unload_cargo=function(squad)
        local p,s,f,o=squad_info(squad)
        local pos=o.position
        local area
        local CDLR=CARGO_DRONE_LOAD_RANGE
        if pos and (pos.x-p.x)^2+(pos.y-p.y)^2>CDLR^2 then 
        --if pos and (((pos.x-p.x)>CARGO_DRONE_LOAD_RANGE) or ((pos.y-p.y)>CARGO_DRONE_LOAD_RANGE)) then 
            return false 
        elseif pos then 
            area={{pos.x-1,pos.y-1,},{pos.x+1,pos.y+1}}
        else --this branch is currently unreachable due to combinator mechanics
            area={{p.x-CDLR,p.y-CDLR},{p.x+CDLR,p.y+CDLR}}
        end
        local chests=s.find_entities_filtered{
            area=area,
            type = "logistic-container",
            force = f
        }
        if #chests==0 then return false end
        
        --print('unload')
        local inventories=get_inventories(chests)
        local drones=check_valid(squad.cargo)
        if #drones==0 then return false end
        --print('drones')
        
        local next_drone=gen_special_drone_iterator(drones,cargo_is_empty)
        local next_inv=gen_special_drone_iterator(inventories,inventory_is_full)
        
        local dr,inv=next_drone(),next_inv()
        
        while dr and inv do
            for name,count in pairs(dr.items) do
                if count and count>0 then
                    local inserted=inv.insert{name=name,count=count}
                    dr.total_count=dr.total_count-inserted
                    dr.items[name]=dr.items[name] - inserted
                    if inserted==0 then break end 
                end
            end
            dr,inv=next_drone(),next_inv()
        end
        --print(serpent.block(drones))
        return true
    end,
    drop_construction=function(squad)
        local p,s,f,o=squad_info(squad)
        local pos=o.position
        
        local drones=valid_construction(squad.construction)
        if not drones or #drones==0 then return false end
        
        if (pos) and ((pos.x-p.x)^2+(pos.y-p.y)^2>CONSTRUCTION_DROP_RANGE^2) then return false end
        --if (pos) and ((pos.x-p.x)>CONSTRUCTION_DROP_RANGE or (pos.y-p.y)>CONSTRUCTION_DROP_RANGE) then return false end
        if pos then
            local drone=remove(drones)
            return spawn_roboport(drone,s,pos,f)
        end
    end,
    drop_construction_on_ghost=function(squad)
        local p,s,f,o=squad_info(squad)
        local pos=o.position
        
        local drones=valid_construction(squad.construction)
        if not drones or #drones==0 then return false end
        
        local CDR=CONSTRUCTION_DROP_RANGE
        if (not pos) or ((pos.x-p.x)^2+(pos.y-p.y)^2>CONSTRUCTION_DROP_RANGE^2) then return false end
        --if (not pos) or ((pos.x-p.x)>CDR) or ((pos.y-p.y)>CDR) then return false end
        
        local area={{pos.x-CDR,pos.y-CDR},{pos.x+CDR,pos.y+CDR}}
        local ghosts=s.find_entities_filtered {area=area, name="entity-ghost", force=f,}
        local succ=false
        for i=1,#ghosts do
            local g=ghosts[i]
            if g.ghost_name==CONSTRUCTION_DROP_RESULT then
                local drone=remove(drones)
                if not drone then return false end
                local success,entity = g.revive()
                succ=succ or success
                if entity then 
                    insert_items(entity)
                    script.raise_event(defines.events.on_robot_built_entity,{
                        robot=drone,
                        created_entity=entity,
                        name=defines.events.on_robot_built_entity,
                        tick=game.tick,
                    })
                    LANDING_PADS.ignore_death(drone)
                    drone.die()
                end
            end
        end
        
        return succ and true
    end,
    load_carrier=function(squad)
        
    end,
    launch_carrier=function(squad)
        
    end,
    unload_carrier=function(squad)
        
    end,
}
return SPECIAL_HANDLERS
