require "my_util"
require"scripts/drones"--set on init
require "scripts/guides"


local DATA--link to global.table

local DRONES,GUIDES
local DRONE_NAMES={
    cargo=my"cargo-drone",
    roboport=my"construction-drone",
    --carrier=my"carrier",
}
local SIGNALS=require"scripts/squad_signals"

local TTL_OFFSET=61

local prototype_parameters=require"data_u/squad"
local DISTRACTION_DURATION=prototype_parameters.DISTRACTION_DURATION--updates (tied to transmitter updates)
local DISTRACTION_RANGE=prototype_parameters.DISTRACTION_RANGE--tiles
local DESTINATION_NAME=prototype_parameters.DESTINATION_NAME
local DISTRACTION_NAME=prototype_parameters.DISTRACTION_NAME

local COMBAT_DURATION=10--updates

local insert=table.insert
local remove=table.remove

local SIGNAL_HANDLERS=require"scripts/special_orders"

local SIGNAL_INDICES={--[[
right now there are two different combinators 
and none of the signals is used in both
but these constants will be split another day
]]
    posX=1,
    posY=2,
    destX=3,
    destY=4,
    distraction=5,
    dest_dist=6,
    home_dist=7,
    done=8,
    time_to_live=9,
    dest_label=10,
    order_name=11,
    homeX=12,
    homeY=13,
    in_combat=14,
    other_start=15,
}
local MAKE_SIGNAL=function(index,name,value)
    return {count=value,signal={type="virtual",name=name},index=index}
end
local MAKE_ITEM_SIGNAL=function(index,name,value)
    return {count=value,signal={type="item",name=name},index=index}
end

local ORDER_TO_SIGNALS=function(order,drones)
    local SI=SIGNAL_INDICES
    local S=SIGNALS
    local p=order.position
    local x,y=p.x,p.y
    local output={
        MAKE_SIGNAL(SI.destX,S.destX,x),
        MAKE_SIGNAL(SI.destY,S.destY,y),
        MAKE_SIGNAL(SI.distraction,S.distraction,order.allow_distraction and 1 or 0),
        MAKE_SIGNAL(SI.dest_label,S.dest_label,label_name or 0),
    }
    if order.name then insert(output,MAKE_SIGNAL(SI.order_name,S[order.name],1)) end
    local i=SI.other_start
    for _,d in pairs(drones) do
        local name,count=d.name,d.count
        insert(output,MAKE_ITEM_SIGNAL(i,DRONES[name].capsule,count))
        i=i+1
    end
    return output
end
 
local sort=table.sort
local compute_median=function(tab)
    sort(tab)
    local c=(1+#tab)/2
    c=c-c%1
    return tab[c] or 0
end

local sqrt=math.sqrt
local SQUAD
SQUAD={
    new=function(transmitter)
        local o={
            valid=false,
            order=false,
            drones={},
            guide=false,
            destination=false,
            transmitter=transmitter,
            cargo={},
            construction={},
            distraction_duration=false,
            in_combat=0,
            output=false,
        }
        return setmetatable(o,SQUAD)
    end,
    restore=function(o)
        setmetatable(o,SQUAD)
    end,
    update=function(o)
        if not (o.valid and o.guide.valid) then
            o:despawn_squad()
            local SI,S=SIGNAL_INDICES,SIGNALS
            local p3=o.transmitter:position()
            local xh,yh=p3.x,p3.y
            return {
                MAKE_SIGNAL(SI.homeX,S.homeX,xh),
                MAKE_SIGNAL(SI.homeY,S.homeY,yh),
            }  
        end
        
        local c=o.distraction_duration
        if c and c>0 then
            o.distraction_duration=c-1
        else
            o:back_on_track()
        end
        
        local combat=(o.in_combat or 0)-1
        combat = combat>0 and combat or 0
        o.in_combat=combat
        
        local order_success=0
        if o.order.name then
            local succ=SIGNAL_HANDLERS[o.order.name](o)
            --print(o.order.name,succ,serpent.block(o.order.position),serpent.block(o.guide.position))
            order_success= (succ and 1) or -1
        end
        
        local d=o.drones
        local alive={}
        local ttl={}
        for i=#d,1,-1 do
            local drone=d[i]
            if not drone.valid then 
                remove(d,i) 
            else
                local c=alive[drone.name] or 0
                alive[drone.name]=c+1
                insert(ttl,drone.time_to_live)
            end
        end
                
        local median_ttl=(compute_median(ttl)-TTL_OFFSET)/60
        
        local p=o.guide.position
        local x,y=p.x,p.y
        local p2=o.order.position
        local x2,y2=p2.x,p2.y
        local dist=sqrt((x2-x)^2+(y2-y)^2)
        local p3=o.transmitter:position()
        local xh,yh=p3.x,p3.y
        local dist_h=sqrt((xh-x)^2+(yh-y)^2)
        local distraction=o.order.allow_distraction and 1 or 0
        local label=o.order.label or 0
        
        local SI=SIGNAL_INDICES
        local S=SIGNALS
        local output={
            MAKE_SIGNAL(SI.posX,S.posX,x),
            MAKE_SIGNAL(SI.posY,S.posY,y),
            MAKE_SIGNAL(SI.homeX,S.homeX,xh),
            MAKE_SIGNAL(SI.homeY,S.homeY,yh),
            MAKE_SIGNAL(SI.dest_dist,S.dest_dist,dist),
            MAKE_SIGNAL(SI.home_dist,S.home_dist,dist_h),
            MAKE_SIGNAL(SI.done,S.done,order_success),
            MAKE_SIGNAL(SI.time_to_live,S.time_to_live,median_ttl),
            MAKE_SIGNAL(SI.in_combat,S.in_combat,combat),
        }
        local i=SI.other_start
        for name,count in pairs(alive) do
            insert(output,MAKE_ITEM_SIGNAL(i,DRONES[name].capsule,count))
            i=i+1
        end
        if #o.drones==0 then o:despawn_squad() end
        return output
    end,
    delete=function(o)
        for i=#DATA,1,-1 do
            if DATA[i]==o then
                o:despawn_squad()
                table.remove(DATA,i) 
                return 
            end
        end
    end,
    process_signals=function(me,input)
        local S=SIGNALS
        local l=input[S.map_label]
        local x,y,label_name
        if l then
            local f_id=me.transmitter:force().name
            local lab=label.get(f_id,l)
            if lab then 
                local p=lab.position
                x,y=p.x,p.y
                label_name=l
            end
        end
        if not (x and y) then
            x,y=input[S.destX],input[S.destY]
        end
        --print(serpent.block(input))
        x=x or 0 
        y=y or 0 --this is factorio fault, there is no difference between signal 0 and absent signal
        local order={position={x=x,y=y,},label=label_name}
        for name,handler in pairs(SIGNAL_HANDLERS) do
            if input[S[name]] then
                order.name=name
                break
            end
        end
        if not (x and y or order.name) then
            return --invalid order
        end
        order.allow_distraction=input[S.distraction] and input[S.distraction]~=0 or false
        
        local to_spawn={}
        local names={}
        local count=0
        for name,drone in pairs(DRONES) do
            local n=input[drone.capsule]
            if n and n>0 then
                insert(to_spawn,{name=name,count=n})
                insert(names,name)
                count=count +n
            end
        end
        if not me.valid and ((count==0) or not me.transmitter:can_spawn(to_spawn)) then return end --no drones to send
        --check whether there are only new robots in squad
        if 0==me:check_alive() then 
            me:despawn_squad() 
            --quick fix to land order at start
            if order.name=='land' then 
                assert(SIGNAL_HANDLERS[order.name])--reminder in case name changes
                return 
            end
        end
        
        if count>0 then
            if not (me.valid and me.guide.valid) then 
                me:spawn_squad(GUIDES:bit_slower(DRONES:slowest(names)))
            end        
            me:spawn_drones(to_spawn) 
        end
        me:set_order(order)
        me.transmitter:update_order_output(ORDER_TO_SIGNALS(order,to_spawn))
    end,
    __index={
        set_order=function(me,order)
            assert( ((order.position.x and order.position.y) or (order.name )))
            me.order=order
        end,
        spawn_squad=function(me,guide_name)
            local p,s,f=me.transmitter:posurforce()
            local d=s.create_entity{name=DESTINATION_NAME,position=p,force="neutral"}
            local g=s.create_entity{name=guide_name,position=p,force=f,target=d}
            d.destructible=false
            g.destructible=false
            me.valid=true
            me.guide=g
            me.destination=d
            DATA[g.unit_number]=me
        end,
        spawn_drones=function(me,to_spawn)
            local drones=me.transmitter:spawn_drones(me.guide,to_spawn)
            local D=DRONES
            local group=me.drones
            local cargo=me.cargo
            local construction=me.construction
            for _,d in pairs(drones) do
                insert(group,d)
                if d.name==DRONE_NAMES.cargo then
                    insert(cargo,{entity=d,items={},total_count=0})
                elseif d.name==DRONE_NAMES.roboport then
                    insert(construction,d)
                --elseif
                
                end
            end
        end,
        check_alive=function(me)
            local d=me.drones
            for i=#d,1,-1 do
                local drone=d[i]
                if not drone.valid then 
                    remove(d,i) 
                end
            end
            return #d
        end,
        kill=function(me) for _,d in pairs(me.drones) do if d.valid then d.die() end end end,
        despawn_squad=function(me)
            me.valid=false
            me.distraction_duration=false
            me.in_combat=0
            local d=me.destination
            if d and d.valid then d.destroy() end
            local g=me.guide
            if g and g.valid then 
                DATA[g.unit_number]=nil
                g.destroy() 
            end
            for _,dr in pairs(me.drones) do dr.die() end
            me.order=false
            me.transmitter:update_order_output()
        end,
        distract=function(me,position)
            me.in_combat=COMBAT_DURATION
            if not me.order.allow_distraction then return end
            me.distraction_duration=DISTRACTION_DURATION
            me.destination.teleport(position)
        end,
        back_on_track=function(me)
            me.distraction_duration=false
            me.destination.teleport(me.order.position)
        end,
        in_combat=function(me)
            return me.in_combat>0 and true or false
        end,
        get_signals=function(me)
            return table.deepcopy(me.output or {})
        end,      
    }
}

on_init(function()
    DATA=new_storage("squads")
    DRONES=get_storage("drone-types")
    GUIDES=get_storage("guide-types")
end)
on_load(function()
    DATA=get_storage("squads")
    for _,o in pairs(DATA) do SQUAD.restore(o) end
    DRONES=get_storage("drone-types")
    GUIDES=get_storage("guide-types")
end)
on_trigger(function(event)
--just pulls every squad in the vicinity instead of true distraction
--might as well get away with that
    local e=event.entity
    if not e.name==DISTRACTION_NAME then return end
    local p1=e.position
    local D2=DISTRACTION_RANGE^2
    for _,s in pairs(DATA) do
        local p2=s.guide.position
        if (p1.x-p2.x)^2+(p1.y-p2.y)^2<D2 then
            s:distract(p1)
        end
    end
end)

return SQUAD
