require "my_util"
require "scripts/drones"
local drone_life=require"scripts/drone_life"

local STEP=60--ticks between update of same entity
local SPAWN_SCATTER=10--radius for spawning the entities from the chests

local random=math.random
local RANDOM_SHIFT=function() return random(0,SPAWN_SCATTER) end

local GREEN,RED=assert(defines.wire_type.green,defines.wire_type.red)

local SIGNALS=require"scripts/squad_signals"

local DRONES

local SQUAD=require"scripts/squad"

local insert=table.insert
local ceil=math.ceil

local handle_connections
do
    local connectable={
        ["arithmetic-combinator"]=defines.circuit_connector_id.combinator_input,
        ["decider-combinator"]=defines.circuit_connector_id.combinator_input,
        ["constant-combinator"]=defines.circuit_connector_id.constant_combinator,
    }
    local areas={
        [my"command-transmitter-output"]={x={-1,1},y={0.5,1.5}},
        [my"command-transmitter-order-output"]={x={0.5,1.5},y={-1,1}},
        [my"command-transmitter-inventory"]={x={-1,1},y={-1.5,-0.5}},
    }
    handle_connections=function(e)
        local s,f,p=e.surface,e.force,e.position
        local ar,id=areas[e.name],connectable[e.name]
        if connectable[e.name] then
            local area={{p.x-1.5,p.y-1.5},{p.x+1.5,p.y+1.5}}
            for _,ent in pairs(s.find_entities_filtered{area=area,force=f}) do
                local ar=areas[ent.name]
                if ar then
                    local pos=ent.position
                    local dx,dy=(p.x-pos.x),(p.y-pos.y)
                    if (ar.x[1]<=dx) and (dx<=ar.x[2]) and (ar.y[1]<=dy) and (dy<=ar.y[2]) then
                        ent.connect_neighbour{
                            target_entity=e,
                            wire=RED,
                            target_circuit_id=id,
                        }
                    end
                end
            end
        elseif ar then
            local x,y=p.x,p.y
            local area={{x+ar.x[1],y+ar.y[1]},{x+ar.x[2],y+ar.y[2]}}
            for _,ent in pairs(s.find_entities_filtered{area=area,force=f}) do
                local id=connectable[ent.name]
                if id then
                    e.connect_neighbour{
                        target_entity=ent,
                        wire=RED,
                        target_circuit_id=id,
                    }
                end
            end
        end
    end
end

local invalid={valid=false}
local TRANSMITTER
TRANSMITTER={
    new=function(e)
        local s,f,p=e.surface,e.force,e.position
        local z=0.5--pos shift
        local out=s.create_entity{name=my"command-transmitter-output",position={x=p.x-z,y=p.y+z},force=f}
        local ord=s.create_entity{name=my"command-transmitter-order-output",position={x=p.x+z,y=p.y+z},force=f}
        local inv=s.create_entity{name=my"command-transmitter-inventory",position={x=p.x+z,y=p.y-z},force=f}
        
        out.destructible=false
        inv.destructible=false
        
        local o={
            input_ent=e,
            output_ent=out,
            order_output_ent=ord,
            chest_ent=inv,
            squad=false,
            --order=false,
            inventory=inv.get_inventory(defines.inventory.chest),
            in_green=invalid,
            in_red=invalid,
            output=out.get_or_create_control_behavior(),
            order_output=ord.get_or_create_control_behavior(),
            --last_output=false,
        }
        o.squad=SQUAD.new(o)
        return setmetatable(o,TRANSMITTER)
    end,
    restore=function(o)
        setmetatable(o,TRANSMITTER)
        SQUAD.restore(o.squad)
    end,
    __index={
        position=function(me)
            return me.input_ent.position
        end,
        can_spawn=function(me,drones)
            if me.input_ent.energy==0 then return false end
            local inventory=me.inventory
            if not inventory.valid then return end
            local D=DRONES
            for _,drn in pairs(drones) do
                local capsule=D[drn.name].capsule
                if inventory.get_item_count(capsule) >0 then return true end
            end
        end,
        spawn_drones=function(me,guide,drones)
            if me.input_ent.energy==0 then return {} end
            local inventory=me.inventory
            if not inventory.valid then return end
            local present_drones={}
            local D=DRONES
            for _,drn in pairs(drones) do
                local name,count=drn.name,drn.count
                local d=D[name]
                local capsule=d.capsule
                local n=d.multiplier
                present_drones[name]=inventory.remove{name=capsule,count=ceil(count/n)}
            end
            local c=me.chest_ent
            local s,f,p=c.surface,c.force,c.position
            local spawned_drones={}
            for name,count in pairs(present_drones) do
                local d=D[name]
                if d then
                    local n=d.multiplier
                    local projectile=d.projectile
                    for i=1,count*n do
                        local dx,dy=RANDOM_SHIFT(),RANDOM_SHIFT()
                        local pos={p.x+dx,p.y+dy}
                        local drone=s.create_entity{
                            name=name,
                            force=f,
                            position=pos,
                            target=guide,
                            }
                        insert(spawned_drones,drone)
                        s.create_entity{name=projectile,position=p,target=drone,speed=0.4}
                    end
                end
            end
            drone_life.register_drones(spawned_drones)
            return spawned_drones
        end,
        force=function(me)
            return me.input_ent.force
        end,
        posurforce=function(me)
            local e=me.input_ent
            return e.position, e.surface, e.force
        end,
        update_order_output=function(me,parameters)
            me.order_output.parameters={parameters=parameters or {}}
        end,
        cleanup=function(me)
            SQUAD.delete(me.squad)
            local e1,e2,e3,e4=me.input_ent,me.output_ent,me.chest_ent,me.order_output_ent
            if e1.valid then e1.destroy() end
            if e2.valid then e2.destroy() end
            if e3.valid then e3.destroy() end
            if e4.valid then e4.destroy() end
        end,
    }
}

local data

on_init(function()
    data=new_storage("sender")
    DRONES=get_storage("drone-types")
end)
on_load(function()
    data=get_storage("sender")
    for i=1,#data do TRANSMITTER.restore(data[i]) end
    DRONES=get_storage("drone-types")
end)

on_build(function(event)
    local e=event.created_entity
    if (e.name==my"command-transmitter-input") then 
        local o=TRANSMITTER.new(e)
        insert(data,o)
        handle_connections(o.chest_ent)
        handle_connections(o.order_output_ent)
        handle_connections(o.output_ent)
    else
        handle_connections(e)
    end
end)

local on_remove=function(event)
    local e=event.entity
    if not (e.name==my"command-transmitter-input") then return end
    local data=data
    for i=1,#data do
        if data[i].input_ent==e then
            table.remove(data,i):cleanup()
            return 
        end
    end
end

on_mined(on_remove)
on_die(on_remove)

local update=function(d)
    local input_ent,output_ent,chest_ent=d.input_ent,d.output_ent,d.chest_ent
    if not (input_ent.valid and output_ent.valid and chest_ent.valid) then
        return true
    end
    --update network connections
    if not d.in_green.valid then d.in_green= input_ent.get_circuit_network(GREEN) or invalid end
    if not d.in_red.valid then d.in_red= input_ent.get_circuit_network(RED) or invalid end
    if not d.output.valid then d.out= output_ent.get_or_create_control_behavior() or invalid end
    
    --get incoming signals
    local in_signals={}
    if d.in_green.valid then
        local green=d.in_green.signals
        if green then for i=1,#green do
            local s=green[i]
        --for ,s in pairs(green) do
            in_signals[s.signal.name]=s.count 
        end end
    end
    if d.in_red.valid then
        local red=d.in_red.signals
        if red then for i=1,#red do
            local s=red[i]
        --for k,v in pairs(r.in_red.signals) do
            local name,count=s.signal.name,s.count
            in_signals[name]=(in_signals[name] or 0)+count
        end end
    end
    --send command to squad
    local send=in_signals[SIGNALS.send]
    if input_ent.energy>0 and send and send>0  then
        SQUAD.process_signals(d.squad,in_signals)
    end

    --update output
    local out_signals=SQUAD.update(d.squad)
    --if out_signals then
        d.output.parameters={parameters=out_signals or {}}
    --end
end

on_tick(function(event)
    local t=event.tick
    for i=#data-t%STEP,1,-STEP do--update each 60th radar each tick
        local delet_this=update(data[i])
        if delet_this then table.remove(data,i):cleanup() end
    end
end)
