--[[script to perform all kinds of tests
it should probably be split, into separate files, 
but there's no way to keep access to locals inside `require`d file
and I rely on locals for closure generation a lot here
So it is just one grand file which needs code folding redactor to be readable
]]
require "my_util"

local not_done_yet=function() print("not implemented")  end

local INDEX={--chapters of testing
    --landing=true,
    --radar=true,
    --label=true,
    transmitter=true,
    --squad=true,
    --manual=true,
}

local PLAYER,PC,PIND, PF, SURF
local E=defines.events
local DATA
local TESTS={}
--local STARTING_N=10

local test_in_progress
do
    local tip=false
    test_in_progress=function(arg)
        if not (arg==nil) then tip=arg end
        return tip
    end
end

local run_tests=function()
    local s=DATA.test_n or STARTING_N or 1
    for i=s,#TESTS do
        if test_in_progress() then return end
        local f=TESTS[i]
        local name=TESTS[f]
        print(name)
        f()
        DATA.test_n=i+1
    end
end

local ensure=function(val)
    if not val then 
        print("Error at",debug.getinfo(2).currentline)
        PLAYER.print("Error at "..debug.getinfo(2).currentline)
        DATA.unsaveable=true
        DATA.test_n=1+ #TESTS
    end
    return val
end

local test=function(name,t)
    assert(type(t)=='function')
    table.insert(TESTS,t)
    TESTS[t]="--<< "..name..">>--"
    TESTS[name]=t
end

on_player_created(function(event)
    run_tests()
end)
on_tick(run_tests)
on_init(function()
    print("TESTING INITIATED")
    DATA=new_storage("tests")
end)
on_load(function()
    DATA=get_storage("tests")
    if DATA.unsaveable then print("tests messed up by save") end
end)
on_trigger(function(e)
    --print(serpent.block(e.entity.position))
end)

local old_handler={}
local additional_handler=function(event,fun)
    ensure(type(event)=='number',"wrong event id: "..(event or "nil"))
    local f=script.get_event_handler(event)
    DATA.unsaveable=true
    old_handler[event]=f
    script.on_event(event,function(e) fun(e) f(e) end)
end
local restore_handler=function(event)
    ensure(type(event)=='number',"wrong event id: "..(event or "nil"))
    script.on_event(event,old_handler[event])
    old_handler[event]=nil
end

local timer=function(n) return function() n=n-1 return n<0 end  end
local wait=function(till,after_wait)
    test_in_progress(true)
    game.speed=30
    additional_handler(E.on_tick,function(e)
        if till() then
            restore_handler(E.on_tick)
            test_in_progress(false)
            game.speed=1
            if after_wait then after_wait() end
        end
    end)
end
local do_nothing=function()  end

do--block for code folding, initial setup
    test('give_items',function() 
        local p=game.players[1]
        p.insert("defender-capsule")
        p.insert("submachine-gun")
        p.insert("piercing-rounds-magazine")
        p.insert"constant-combinator"
        p.force.research_all_technologies()
    end)

    test('init common variables',function()
        PLAYER=game.players[1]
        PC=PLAYER.character
        PIND=PLAYER.index
        PF=game.forces.player
        SURF=game.surfaces.nauvis
        SURF.always_day=true
    end)
end

local spawn_drone= function(pos) 
    local drone=SURF.create_entity{name="defender",force=PF, position=pos or {0,0},target=PC} 
    script.raise_event(
        E.on_trigger_created_entity,
        {tick=game.tick, name=E.on_trigger_created_entity, entity=drone}
    )
    return drone
end

local build=function(name,pos,force)
    --print(debug.traceback())
    local ent=SURF.create_entity{name=name, force=force or PF, position=pos or {0,0}}
    script.raise_event(
        E.on_built_entity,
        {tick=game.tick, name=E.on_built_entity, player_index=PIND, created_entity=ent}
    )
    return ent
end

local build_rebuild=function(ent,name,pos)
    return ent and ent.valid and ent or build(name,pos)
end

local setup_power=function()
    DATA.substation = build_rebuild(DATA.substation,"substation",{0,7})
    local solars = DATA.power or {}
    local x=-7
    for i=1,5 do
        local s=solars[i]
        solars[i]= build_rebuild(s,'solar-panel',{x,7})
        x=x+3
    end
    DATA.power=solars
end

local setup_roboport=function()
    DATA.roboport = build_rebuild( DATA.roboport,"roboport",{0,13})
    DATA.roboport.insert('construction-robot')
    DATA.provider= build_rebuild( DATA.provider, "logistic-chest-passive-provider", {5,13})
end

local params2dict=function(parameters)
    local res={}
    --print(serpent.block(parameters.parameters))
    for _,s in pairs(parameters.parameters) do
        if s.signal.name then
            res[s.signal.name]=s.count
        end
    end
    return res
end
--drone landing testing
if INDEX.landing then
    test('no landing pads landing',function() spawn_drone().die() end)


    test('build basic landing pad',function()  
        local pad=build(my"basic-landing-pad",{1,0})
        DATA.pad= pad
        DATA.inv= pad.get_inventory(defines.inventory.chest)
    end)


    test('drone landing on basic',function() 
        local i=DATA.inv
        i.clear()
        spawn_drone().die()
        ensure(i.get_item_count()>0)
    end)

    test('unrelated death',function() build("gun-turret",{0,5}).die() end)

    test('dying statistics',function()
        local inv=DATA.inv
        inv.clear()
        for i=1,100 do spawn_drone().die() end
        for i=1,100 do spawn_drone().damage(300,PF) end
        ensure(inv.get_item_count(my"#wrecked#-defender")>0)
        ensure(inv.get_item_count(my"#spent#-defender")>0)
    end)

    test('drone life',function()
        spawn_drone()
        game.speed=20
        local inv=DATA.inv
        inv.clear()
        test_in_progress(true)
        additional_handler(E.on_entity_died,function(e)
            if not (e.entity.name=='defender') then return end
            restore_handler(E.on_entity_died)
            additional_handler(E.on_tick,function()  
                restore_handler(E.on_tick)
                ensure(DATA.inv.get_item_count()>0)
                game.speed=1
                test_in_progress(not true)
            end)
        end)
    end)

    test('failed landing',function()
        local inv=DATA.inv
        inv.clear()
        local n='iron-plate'
        repeat 
            local r= inv.insert(n)
        until (r==0)
        local c=inv.get_item_count(n)
        spawn_drone().die()
        ensure(inv.get_item_count(n)==c)
    end)

    test('1 drone 2 pads',function()
        local pad2=build(my"basic-landing-pad",{-1,0})
        DATA.pad2=pad2
        local inv2=pad2.get_inventory(defines.inventory.chest)
        DATA.inv2=inv2
        spawn_drone().die()
        ensure(inv2.get_item_count()>0)
        inv2.clear()
        local inv = DATA.inv
        inv.clear()
        spawn_drone({-10.5,0}).die()
        ensure(inv2.get_item_count()>0)
        inv2.clear()
        spawn_drone().die()
        ensure(inv2.get_item_count()>0)
        inv2.clear() inv.clear()
        repeat 
            local r= inv2.insert('iron-plate')
        until (r==0)
        spawn_drone().die()
        ensure(inv.get_item_count()>0)
    end)

    test("Landing outside of range",function()
        local inv=DATA.inv
        inv.clear()
        ensure(inv.get_item_count()==0)
        spawn_drone({40,0}).die()
        ensure(inv.get_item_count()==0)
    end)

    test("ded pad",function()  
        DATA.inv2.clear()
        DATA.pad.die()
        spawn_drone().die()
        ensure(DATA.inv2.get_item_count()>0)
        DATA.pad2.die()
    end)

    test("powered building",function()
        setup_power()
        build (my"powered-landing-pad-power",{0,1})
    end)
    test("power landing",function()
        game.speed=30
        wait(timer(60*60),
            function()
                local ppad= LANDING_PADS.find({0,0},game.surfaces.nauvis,game.forces.player)
                ensure(ppad.entity.valid and ppad.power_entity.valid)
                DATA.powered=ppad
                DATA.pinv=ppad.entity.get_inventory(defines.inventory.chest)
                local inv=DATA.pinv
                local power= ppad.power_entity
                power.energy=0
                spawn_drone({0,2}).die()
                ensure(inv.get_item_count()==0)
                wait(
                    function() return power.energy>2e4 end,
                    function()
                        game.speed=1
                        local c=power.energy
                        spawn_drone({0,10.9}).die()
                        ensure(inv.get_item_count()>0)
                        ensure(power.energy<c)
                        wait(timer(2))
                    end
                )
            end
        )
    end)
    test("power die",function() DATA.powered.entity.die() spawn_drone().die() end)
end
--signals raw obtaining and testing
--the code below needs testing itself
local signals={map_label=my"map-label"}
local position_signal
test("signal-table",function()
    for n,t in pairs(require"data_u/signals") do
        if n:find("radar") then
            local pat=my"radar%-(.*)%"
            for _,s in pairs(t) do
                signals[s:match(pat):gsub("%-","_")]=s
            end
            break
        end
    end
    position_signal=function(x,y)
        return {
            {signal={name=signals.posX,type='virtual'},count=x,index=1},
            {signal={name=signals.posY,type='virtual'},count=y or 1,index=2}
        }
    end
end)
--smart radar testing
if INDEX.radar then
    test("smart-radar",function()
        setup_power()
        local rin = build(my"smart-radar-input",{3, 0})
        local rout = SURF.find_entities_filtered{name= my"smart-radar-output"}[1]
        DATA.radar_in=rin
        DATA.radar_out=rout
        DATA.radar_out_interface=rout.get_control_behavior()
        ensure(rin.valid)
        ensure(rout.valid)
        DATA.radar_in.energy=1e9
        local inputter=build("constant-combinator",{3,3})
        inputter.connect_neighbour{target_entity=rin, wire=defines.wire_type.red}
        DATA.inputter=inputter
        DATA.iface=inputter.get_control_behavior()
    end)

    local later=function(do_smth)
        return function()
            game.speed=30
            wait(timer(120),function() game.speed=1 do_smth() end)
        end
    end

    local react_to_signals = function(signals, do_smth)
        local cc=DATA.iface
        cc.parameters={parameters=signals}
        later(do_smth)()
    end

    local get_output=function()
        local out= DATA.radar_out_interface
        return params2dict(out.parameters)
    end

    test("radar signals",function()
        ensure(signals.posX)
        ensure(signals.posY)
        ensure(signals.force_own)
    end)
    
    test("green wire",function()  
        local cc=build("constant-combinator",{4,3})
        cc.connect_neighbour{target_entity=DATA.radar_in, wire=defines.wire_type.green}
        wait(timer(120),function()  
            cc.get_control_behavior().parameters={parameters=position_signal(1,1)}
            wait(timer(120),function()
                local o=get_output()
                local t=o[signals.time_remaining]*60
                wait(timer(t+3),function()  
                    ensure(get_output()[signals.posX]==1)
                    cc.die()
                end)
            end)
        end)
    end)
    
    test("scan",function()
        local s=signals
        local chest=build("iron-chest",{16,16})
        DATA.chest=chest
        react_to_signals(position_signal(1,1),
        function ()
            local out= DATA.radar_out_interface
            local res=params2dict(out.parameters)
            if not ensure(res[s.time_remaining]) then return end
            DATA.radar_in.energy=1e9
            local e=DATA.radar_in.energy
            wait(timer(60*res[s.time_remaining]+10),
                function()
                    local res=params2dict(out.parameters)
                    ensure(res[s.posX] and res[s.posX]==1)
                    ensure(res[s.posY] and res[s.posY]==1)
                    ensure(res[chest.name])
                    ensure(DATA.radar_in.energy<e)
                end
            )
        end)
    end)

    test('out of range scan',later(function()
        local s=signals
        react_to_signals(position_signal(21*32,1),
        function()
            local res=get_output()
            ensure(res[s.out_of_range])
        end
        )
    end))

    test("scan after error",later(TESTS["scan"]))

    test("out of range again",later(TESTS['out of range scan']))

    test("no power scan",later(function()
        DATA.substation.die()
        DATA.radar_in.energy=1
        react_to_signals(position_signal(1,1),
        function ()
            game.speed=30
            wait(timer(60*62),function ()
                game.speed=1
                local res=get_output()
                local chest=DATA.chest
                ensure( not res[chest.name])
                DATA.substation=build('substation',{0,7})
            end)
        end)
    end))

    test("radar death",function()
        DATA.radar_in.die()
        wait(timer(61),function() ensure(not DATA.radar_out.valid)  end)
    end)
    
    test("radar rebuild",function()
        setup_roboport()
        DATA.provider.insert{name=my'smart-radar-item',count=1}
        wait(timer(60*20),function()
            local r=ensure(SURF.find_entities_filtered{name= my"smart-radar-output"}[1])
            ensure(r.valid)
            r.die() 
            ensure(not r.valid)
        end)
    end)
end
--label testing
if INDEX.label then
    local spawn_gui=function()
        local n=my"map-label-gui"
        local h=script.get_event_handler(n)
        --script.raise_event(n,{name=n,player_index=PIND,tick=game.tick})
        h({name=n,player_index=PIND,tick=game.tick})
    end
    
    local get_gui=function() return ensure(PLAYER.gui.center["label-gui"])  end
    
    test("label-gui-spawn",function()
        spawn_gui()
        ensure(get_gui())
    end)
    
    local function find_in_tree(node,next_fun,done_fun)
        --print(node.name)
        if done_fun(node) then return node end
        for _,n in pairs(next_fun(node)) do
            local res= find_in_tree(n,next_fun,done_fun)
            if res then return res end
        end
    end
    local children=function(gui_el) return (gui_el.children) or {} end
    local find_text_field=function(gui_el,name)
        local is_text=function(gui_el) if gui_el.type=='textfield' and gui_el.name==name then return gui_el end end
        return find_in_tree(gui_el,children,is_text)
    end
    local find_button=function(gui_el,name)
        return find_in_tree(gui_el,children,function(el) return el.type=="button" and el.name==name and el end)
    end
    local click=function(element)
        script.raise_event(defines.events.on_gui_click,{name=defines.events.on_gui_click,
            element=element,
            player_index=element.player_index,
            button=defines.mouse_button_type.left,
        })
    end
    local get_xyn=function(g)
        return ensure(find_text_field(g,"x")),ensure(find_text_field(g,"y")),ensure(find_text_field(g,"nametext"))
    end
    local set_xyn=function(g,trio)  
        local x,y,n=get_xyn(g)
        x.text,y.text,n.text=table.unpack(trio)
    end
    local triples={
        {-10,-12,1},{14,0,2},{20,14,3},{134,12,8},{-213,12,32}
    }
    test("label-gui-button-clicks",function ()
        spawn_gui()
        local g=get_gui()
        for _,tr in ipairs(triples) do
            set_xyn(g,tr)
            click(find_button(g,"set"))
        end
        --ensure(#SURF.find_entities_filtered{name=LABEL_ENTITY_NAME}==2)--entities are not created currently
        click(find_button(g,"close"))
    end)
    
    test("prev button at origin",function()
        spawn_gui()
        click(ensure(find_button(get_gui(),"previous")))
    end)
    test("next prev buttons",function()
        spawn_gui()
        local g=get_gui()
        for i=1,6 do click(ensure(find_button(g,'next'))) end
        for i=1,8 do click(ensure(find_button(g,'previous'))) end
    end)
    test("next prev detailed",function()
        spawn_gui()
        local g=get_gui()
        local x,y,name=get_xyn(g)
        ensure(name.text=="")
        local next=find_button(g,"next")
        for _,trio in ipairs(triples) do
            click(next)
            local out_trio=table.pack(get_xyn(g))
            for i=1,#trio do
                if not ensure(tonumber(out_trio[i].text)==trio[i]) then return end
            end
        end
    end)
    test("label remove",function() 
        spawn_gui()
        local g=get_gui()
        local deleted=select(3,get_xyn(g)).text
        click(find_button(g,"remove"))
        local next=find_button(g,"next")
        for i=1,#triples do
            click(next)
            local n=select(3,get_xyn(g)).text
            ensure(not (n==deleted))
        end
    end)
    test("prev button at origin after removal",function()
        spawn_gui()
        click(ensure(find_button(get_gui(),"previous")))
    end)
    test("closing-gui",function() click(ensure(find_button(get_gui(),"close")))  end)
end    
if INDEX.label and INDEX.radar then--cross functionality
    test("labels to radars",function()
        TESTS["smart-radar"]()
        TESTS["label-gui-button-clicks"]()
        local input=DATA.iface
        local output=DATA.radar_out_interface
        input.parameters={parameters={{signal={name=signals.map_label,type='virtual'},count=1,index=1}}}
        wait(timer(60*32),function()  
            local res=params2dict(output.parameters)
            ensure(res[signals.posX]==-10)
        end)
    end)
    --test("closing gui 2",function () TESTS['closing-gui']() end)
end

if INDEX.transmitter then
    test("clean",function ()
        for _,e in pairs(SURF.find_entities_filtered{}) do 
            if not (e==PC) and e.health then e.die() end
        end
    end)
    
    
    local drone_item="protector-Adil-capsule"
    local drone_name="protector-Adil"
    local signal=function(name,count,index) return {index=index or 1, count=count or 1, signal={type='virtual',name=name}} end
    
    local trans_signals=require"scripts/squad_signals"
    
    test('transmitter-signals',function()
        ensure(trans_signals.map_label)
        ensure(trans_signals.send)
        ensure(trans_signals.done)
    end)
    
    test('transmitter construction',function()
        print("should be redone with order output entity")
        build(my"command-transmitter-input",{0,-7})
        local input_ent=ensure(SURF.find_entities_filtered{name=my"command-transmitter-input"}[1])
        local output_ent=ensure(SURF.find_entities_filtered{name=my"command-transmitter-output"}[1])
        local chest_ent=ensure(SURF.find_entities_filtered{name=my"command-transmitter-inventory"}[1])
        local combinator=build("constant-combinator",{3,-7})
        local input=combinator.get_control_behavior()
        local output=output_ent.get_or_create_control_behavior()
        local inverter=build('arithmetic-combinator')
        local power=build('substation',{-3,-4})
        local tran={input_ent=input_ent, output_ent=output_ent, chest_ent=chest_ent, output=output, input=input, combinator=combinator,inverter=inverter}
        
        DATA.transmitter=tran
        
        combinator.connect_neighbour{target_entity=input_ent, wire=defines.wire_type.red}
        combinator.connect_neighbour{target_entity=inverter, wire=defines.wire_type.red, target_circuit_id=defines.circuit_connector_id.combinator_output}
        output_ent.connect_neighbour{target_entity=inverter, wire=defines.wire_type.red, target_circuit_id=defines.circuit_connector_id.combinator_input}
        inverter.get_control_behavior().parameters={parameters={constant=-1,operation="*",first_signal={type="virtual",name="signal-each"},output_signal={type="item",name=drone_item}}}
        setup_power()
    end)
    
    test("transmitter_rotation",not_done_yet)
    
    local signals_to_transmitter = function(signals, do_smth)
        DATA.transmitter.input.parameters={parameters=signals}
        if do_smth then wait(timer(61),do_smth) end
    end
    
    local clear_input=function()
        signals_to_transmitter{}
    end
    
    local output_from_transmitter=function()
        return params2dict(DATA.transmitter.output.parameters)
    end
    
    local create_squad=function(capsule,count,do_later)
        local ts=trans_signals
        assert(capsule and count)
        signals_to_transmitter({
            signal(ts.destX,0,1),
            signal(ts.destY,0,2),
            signal(ts.send,1,4),
            {index=5,count=count,signal={type='item',name=capsule}}
            },
            do_later
        )
    end

    local kill_drones=function()
        for _,e in pairs(SURF.find_entities_filtered{type="combat-robot"}) do
            e.die()
        end
    end

    test('transmitter-unrelated-signal',function()
        signals_to_transmitter({signal'signal-A'},do_nothing)
    end)
    
    test('squad creation',function()
        local chest=DATA.transmitter.chest_ent
        chest.insert(drone_item)
        local ts=trans_signals
        local inv=chest.get_inventory(defines.inventory.chest)
        local c=inv.get_item_count()
        local count=2
        create_squad(drone_item,count,
            function()
                ensure(#SURF.find_entities_filtered{name=drone_name}==count)
                ensure(c-inv.get_item_count()==count)
                ensure(#SURF.find_entities_filtered{name=my"slow-squad"}==1)
                local s=params2dict(DATA.transmitter.output.parameters)
                local ts=trans_signals
                ensure(s[ts.posX])
                ensure(s[drone_item]==count)
                ensure(s[ts.time_to_live] and s[ts.time_to_live]>200)
            end
        )
    end)
    
    test('squad output signals',function()
        local x=10
        local ts=trans_signals
        signals_to_transmitter({
            signal(ts.destX,x,1),
            signal(ts.destY,0,2),
            signal(ts.send,1,4),
            },
            function()
                local s=params2dict(DATA.transmitter.output.parameters)
                local ts=trans_signals
                ensure((s[ts.destX] or 0)==x)
                ensure(s[ts.posX]-s[ts.destX]<s[ts.dest_dist])
            end
        )
    end)
    
    test('fly around',function()
        local route={{10,12},{-20,1},{100,9},{0,0},{3,200},{0,0}}
        local ts=trans_signals
        local traverse
        traverse=function(i)
            local i,p=next(route,i)
            if not i then return end
            local wait_arrival
            wait_arrival=function()
                local out=DATA.transmitter.output.parameters
                local s=params2dict(out)
                if s[ts.dest_dist]<5 then
                    traverse(i)
                else
                    ensure((s[ts.destX] or 0)==p[1])
                    ensure((s[ts.destY] or 0)==p[2])
                    wait(timer(60*5),wait_arrival)
                end
            end
            
            signals_to_transmitter(
                {signal(ts.destX,p[1],1),signal(ts.destY,p[2],2),signal(ts.send,1,4)},
                wait_arrival
            )
        end
        traverse()
    end)
    
    test("timeout",function()  
        local drones=SURF.find_entities_filtered{name=drone_name}
        local t=drones[1].time_to_live
        local ts=trans_signals
        wait(timer(t+10),function() 
            for _,v in pairs(drones) do ensure(not v.valid)  end 
            local out=DATA.transmitter.output.parameters
            local s=params2dict(out)
            ensure(s[ts.time_to_live]==0)
            ensure(not s[drone_item])  
            wait(timer(120),function() ensure(not next(params2dict(DATA.transmitter.output.parameters))) end)
        end)
    end)
    
    test("green wire",function()  
        local t=DATA.transmitter
        local combinator=t.combinator
        local input_ent=t.input_ent
        local inverter=t.inverter
        local output_ent=t.output_ent
        combinator.disconnect_neighbour(defines.wire_type.red)
        inverter.disconnect_neighbour(defines.wire_type.red)
        combinator.connect_neighbour{target_entity=input_ent, wire=defines.wire_type.green}
        combinator.connect_neighbour{target_entity=inverter, wire=defines.wire_type.green, target_circuit_id=defines.circuit_connector_id.combinator_output}
        output_ent.connect_neighbour{target_entity=inverter, wire=defines.wire_type.green, target_circuit_id=defines.circuit_connector_id.combinator_input}
    end)
    test('green squad creation',TESTS['squad creation'])
    test('green fly around',TESTS['fly around'])
    test('green timeout',TESTS['timeout'])
    
    test("distraction",function()
        local ts=trans_signals
        create_squad(drone_item,6,function ()
            signals_to_transmitter(
                {signal(ts.destX,200,1),signal(ts.destY,0,2),signal(ts.distraction,1,3),signal(ts.send,1,4)},
                function()
                    wait(timer(350),function()
                        local s=params2dict(DATA.transmitter.output.parameters)
                        local enemy=build("biter-spawner",{(s.posX or 0)+20,(s.posY or 0)+28},'enemy')
                        local h=enemy.health
                        wait(timer(600),function()
                            ensure(not enemy.valid or enemy.health<h)
                            local s=params2dict(DATA.transmitter.output.parameters)
                            local ts=trans_signals
                            ensure(s[ts.destX]==200)
                            wait(timer(60*30),function()
                                local s=params2dict(DATA.transmitter.output.parameters)
                                ensure(s[ts.posX]-200<10)
                            end)
                        end)
                    end)
                end
            ) 
        end)
    end)
    
    local order_is_done=function(time)
        return function()
            time=time-1
            ensure(time>0)
            local s=params2dict(DATA.transmitter.output.parameters)
            return s[trans_signals.done]
        end
    end
    
    test("cargo-carry",function ()
        kill_drones()
        local capsule=my("cargo-drone").."-capsule"
        DATA.transmitter.chest_ent.insert(capsule)
        local x,y=3,3
        DATA.chest=assert(build("logistic-chest-passive-provider",{x,y}))
        local ch=DATA.chest.get_inventory(defines.inventory.chest)
        ch.insert("iron-plate")
        local N=ch.get_item_count()
        create_squad(capsule,2,function ()
            local s=params2dict(DATA.transmitter.output.parameters)
            ensure(s[capsule]==2)
            local ts=trans_signals
            local time=600
            signals_to_transmitter(
                {
                    signal(ts.destX,x,1),
                    signal(ts.destY,y,2),
                    signal(ts.load_cargo,1,3),
                    signal(ts.send,1,4)
                },
                function()
                    wait(order_is_done(600),
                    function ()
                        ensure(N>ch.get_item_count())
                        signals_to_transmitter({
                                signal(ts.destX,x,1),
                                signal(ts.destY,y,2),
                                signal(ts.unload_cargo,1,3),
                                signal(ts.send,1,4)
                            },
                            function()
                                wait(order_is_done(600),function () ensure(ch.get_item_count()==N) end)
                            end
                        )
                    end)
                end
            )
        end)
    end)
    test("cargo-carry-full-chest",not_done_yet)
    
    test("cargo-drop",function() 
        local s=params2dict(DATA.transmitter.output.parameters)
        local capsule=my("cargo-drone").."-capsule"
        local n=s[capsule]
        ensure(n>0)
        local ch=DATA.chest.get_inventory(defines.inventory.chest)
        local N=ch.get_item_count()
        ensure(N>0)
        local ts=trans_signals
        local p=DATA.chest.position
        local x,y=8,2
        signals_to_transmitter({
                signal(ts.destX,p.x,1),
                signal(ts.destY,p.y,2),
                signal(ts.load_cargo,1,3),
                signal(ts.send,1,4)
            },
            wait(timer(120),function()
                signals_to_transmitter({
                    signal(ts.destX,x,1),
                    signal(ts.destY,y,2),
                    signal(ts.drop_cargo,1,3),
                    signal(ts.send,1,4)
                },
                function() wait(order_is_done(600),function ()
                    local s=params2dict(DATA.transmitter.output.parameters)
                    ensure(n>s[capsule])
                    local entities=SURF.find_entities_filtered{type="logistic-container",area={{x-1,y-1},{x+1,y+1}}}
                    ensure(#entities==1)
                    local e=entities[1]
                    local inv=e.get_inventory(defines.inventory.chest)
                    ensure(inv.get_item_count()==N)
                end)end
                )
            end)
        )
    end)
    
    local roboport_capsule=my("construction-drone").."-capsule"
    
    local roboport_order=function(order)
        create_squad(roboport_capsule,1,function()
        ensure(output_from_transmitter()[roboport_capsule]==1)
        local ts=trans_signals
        signals_to_transmitter(
            {signal(ts.destX,20,1),signal(ts.destY,1,2),signal(ts.send,1,3)},
            function() wait ( timer(600), function ()
                signals_to_transmitter(
                    {
                    signal(ts.destX,20,1),
                    signal(ts.destY,1,2),
                    signal(ts.send,1,3),
                    signal(order,1,4),
                    },
                    function()
                        local rps=SURF.find_entities_filtered{type="roboport"}
                        ensure(#rps==1)
                        local out=output_from_transmitter()
                        ensure((out[roboport_capsule] or 0)==0)
                        ensure(out[ts.done]==1)
                        rps[1].die()
                    end
                )
            end) end
        )
        end)
    end
    
    test("building-orders",function()  
        local inv=DATA.transmitter.chest_ent
        inv.insert(roboport_capsule)
        roboport_order(trans_signals.drop_construction)
    end)
    
    test("building-orders-blueprint",function()  
        roboport_order(trans_signals.drop_construction_on_ghost)
    end)
    
    test("building-orders-with-blueprint",function()  end)
    
    if INDEX.landing then
        test('squad creation2',function()
            kill_drones()
            local chest=DATA.transmitter.chest_ent
            local item='defender-capsule'
            chest.insert(item)
            local ts=trans_signals
            local inv=chest.get_inventory(defines.inventory.chest)
            local c=inv.get_item_count(item)
            local count=7
            
            DATA.transmitter.inverter.get_control_behavior().parameters={parameters={constant=-1,operation="*",first_signal={type="virtual",name="signal-each"},output_signal={type="item",name=item}}}
            
            create_squad(item,count,function()
                ensure(#SURF.find_entities_filtered{name='defender'}==count)
                ensure(c-inv.get_item_count(item)==count)
                ensure(#SURF.find_entities_filtered{name=my"fast-squad"}==1)
            end)
        end)
        test('landing order',function()  
            local pad=build(my"basic-landing-pad",{1,0})
            local ts=trans_signals
            DATA.landing={pad= pad,inventory= pad.get_inventory(defines.inventory.chest)}
            signals_to_transmitter({
                signal(ts.land,1,3),
                signal(ts.send,1,4),
                },
                function()
                    ensure(#SURF.find_entities_filtered{name='defender'}==0)
                    ensure(DATA.landing.inventory.get_item_count()>0)
                end
            )
        end)
    end
    if INDEX.label then
        test("labels to squad",function()
            local ts=trans_signals
            create_squad(drone_item,6,function()
                signals_to_transmitter({signal(ts.map_label,2,2),signal(ts.send,1,4)},
                    function()
                        local s=params2dict(DATA.transmitter.output.parameters)
                        --label = {14,0,2}
                        ensure(s[ts.destX] ==14)
                        ensure(s[ts.destY] ==0)
                        ensure(s[ts.dest_label]==2)
                    end
                )
            end)
        end)
    end
end

if INDEX.landing and INDEX.radar and INDEX.label and INDEX.transmitter and INDEX.squad then
    test("integral_test",not_done_yet)
end

if INDEX.manual then
    test("look at protector", function() 
        SURF.create_entity{name=my"protector",force=PF,position={0,0}}
    end)
end
test("tests done",do_nothing)
